﻿using System;
using System.Diagnostics;
using System.Net;
using System.ComponentModel;
using System.Threading;
using System.Collections.Generic;
using System.Windows;

namespace SmartBackgroundWorker
{
  /// <summary>
  /// Enumeration for the two execution contexts.
  /// </summary>
  public enum SwitchTo
  {
    /// <summary>
    /// Execution context is UI Thread.
    /// </summary>
    UIThread,

    /// <summary>
    /// Execution context is Worker Thread.
    /// </summary>
    BackgroundThread
  }

  /// <summary>
  /// A smarter background worker than BackgroundWorker for WPF and Silverlight.
  /// </summary>
  public class SmartBackgroundWorker
  {
    /// <summary>
    /// Initializes a new instance of the <see cref="SmartBackgroundWorker"/> class.
    /// </summary>
    public SmartBackgroundWorker()
    {
      if (!Deployment.Current.CheckAccess())
        throw new InvalidOperationException("SmartBackgroundWorker can only be created in the UI thread.");
      _uiSynchronizationContext = SynchronizationContext.Current;
    }

    /// <summary>
    /// SynchronizationContext of UI thread.
    /// </summary>
    readonly SynchronizationContext _uiSynchronizationContext;

    /// <summary>
    /// Starts running a background operation.
    /// </summary>
    public void RunWorkerAsync(IEnumerator<SwitchTo> enumerator)
    {
      lock (Deployment.Current)
      {
        if (_isRunning)
          throw new InvalidOperationException("BackgroundWorker is already running.");

        _isRunning = true;
      }
      ThreadPool.QueueUserWorkItem(OnRun, enumerator);
    }

    /// <summary>
    /// Gets a value indicating whether this instance is busy.
    /// </summary>
    public bool IsBusy
    {
      get { return _isRunning; }
    }

    /// <summary>
    /// True if stepping through worker function.
    /// </summary>
    bool _isRunning;

    /// <summary>
    /// Steps through the worker function.
    /// </summary>
    void OnRun(object argument)
    {
      try
      {
        SwitchTo context = SwitchTo.BackgroundThread;
        var enumerator = (IEnumerator<SwitchTo>)argument;
        bool moveNext = true;
        SendOrPostCallback nextStep = obj =>
                                        {
                                          moveNext = enumerator.MoveNext();
                                          if (moveNext)
                                            context = enumerator.Current;
                                        };
        while (moveNext)
        {
          if (context == SwitchTo.UIThread)
          {
            // Run next step synchronously on UI thread
            _uiSynchronizationContext.Send(nextStep, null);
          }
          else
          {
            // Run next step on background thread
            nextStep(null);
          }
        }
      }
      finally
      {
        _isRunning = false;
      }
    }
  }
}